import asyncio
import contextlib
import datetime
import functools
import logging
import shutil
import typing
from pathlib import Path
from types import TracebackType
from typing import Optional, Self, Type

from gwproto import HardwareLayout, MQTTTopic
from gwproto.messages import CommEvent
from pydantic_settings import BaseSettings

from gwproactor import AppSettings, Proactor, setup_logging
from gwproactor.app import App
from gwproactor.config import DEFAULT_BASE_NAME as DEFAULT_LOG_BASE_NAME
from gwproactor.config import MQTTClient, Paths
from gwproactor_test.certs import copy_keys, uses_tls
from gwproactor_test.clean import hardware_layout_test_path
from gwproactor_test.dummies.pair.child import DummyChildApp
from gwproactor_test.dummies.pair.parent import DummyParentApp
from gwproactor_test.event_consistency_checks import EventAckCounts
from gwproactor_test.instrumented_proactor import (
    InstrumentedProactor,
    MinRangeTuple,
    RangeTuple,
    as_min_range_tuple,
    as_range_tuple,
    caller_str,
    range_min,
)
from gwproactor_test.logger_guard import LoggerGuards
from gwproactor_test.wait import (
    AwaitablePredicate,
    ErrorStringFunction,
    Predicate,
    await_for,
)


def get_option_value(
    *,
    parameter_value: Optional[bool],
    option_name: str,
    request: Optional[typing.Any],
) -> bool:
    if parameter_value is not None or request is None:
        return bool(parameter_value)
    return bool(request.config.getoption(option_name))


class LiveTest:
    _parent_app: App
    _child_app: App
    verbose: bool
    message_summary: bool
    child_verbose: bool
    child_message_summary: bool
    parent_verbose: bool
    parent_message_summary: bool
    parent_on_screen: bool
    lifecycle_logging: bool
    logger_guards: LoggerGuards

    def __init__(
        self,
        *,
        child_app_settings: Optional[AppSettings] = None,
        parent_app_settings: Optional[AppSettings] = None,
        child_layout: Optional[HardwareLayout] = None,
        parent_layout: Optional[HardwareLayout] = None,
        layout: Optional[HardwareLayout] = None,
        verbose: Optional[bool] = None,
        message_summary: Optional[bool] = None,
        child_verbose: Optional[bool] = None,
        child_message_summary: Optional[bool] = None,
        parent_verbose: Optional[bool] = None,
        parent_message_summary: Optional[bool] = None,
        lifecycle_logging: bool = False,
        parent_on_screen: Optional[bool] = None,
        add_child: bool = False,
        add_parent: bool = False,
        add_all: bool = False,
        start_child: bool = False,
        start_parent: bool = False,
        start_all: bool = False,
        request: typing.Any = None,
    ) -> None:
        self.verbose = get_option_value(
            parameter_value=verbose,
            option_name="--live-test-verbose",
            request=request,
        )
        self.message_summary = get_option_value(
            parameter_value=message_summary,
            option_name="--live-test-message-summary",
            request=request,
        )
        self.child_verbose = get_option_value(
            parameter_value=child_verbose,
            option_name="--child-verbose",
            request=request,
        )
        self.child_message_summary = get_option_value(
            parameter_value=child_message_summary,
            option_name="--child-message-summary",
            request=request,
        )
        self.parent_verbose = get_option_value(
            parameter_value=parent_verbose,
            option_name="--parent-verbose",
            request=request,
        )
        self.parent_message_summary = get_option_value(
            parameter_value=parent_message_summary,
            option_name="--parent-message-summary",
            request=request,
        )
        self.parent_on_screen = get_option_value(
            parameter_value=parent_on_screen,
            option_name="--parent-on-screen",
            request=request,
        )
        self.lifecycle_logging = lifecycle_logging
        self._child_app = self._make_app(
            self.child_app_type(),
            child_app_settings,
            app_verbose=self.child_verbose,
            app_message_summary=self.child_message_summary,
            layout=child_layout if child_layout is not None else layout,
        )
        self._parent_app = self._make_app(
            self.parent_app_type(),
            parent_app_settings,
            app_verbose=self.parent_verbose,
            app_message_summary=self.parent_message_summary,
            layout=parent_layout if parent_layout is not None else layout,
        )
        self.setup_logging()
        add_child = add_child or start_child or add_all or start_all
        start_child = start_child or start_all
        if add_child or start_child:
            self.add_child()
            if start_child:
                self.start_child()
        add_parent = add_parent or start_parent or add_all or start_all
        start_parent = start_parent or start_all
        if add_parent or start_parent:
            self.add_parent()
            if start_parent:
                self.start_parent()

    @classmethod
    def child_app_type(cls) -> type[App]:
        return DummyChildApp

    @property
    def child_app(self) -> App:
        return self._child_app

    @classmethod
    def parent_app_type(cls) -> type[App]:
        return DummyParentApp

    @property
    def parent_app(self) -> App:
        return self._parent_app

    @classmethod
    def test_layout_path(cls) -> Path:
        return hardware_layout_test_path()

    def _make_app(
        self,
        app_type: type[App],
        app_settings: Optional[AppSettings],
        *,
        app_verbose: bool = False,
        app_message_summary: bool = False,
        layout: Optional[HardwareLayout] = None,
    ) -> App:
        # Copy hardware layout file.
        paths = Paths(name=app_type.paths_name())
        paths.mkdirs(parents=True, exist_ok=True)
        shutil.copyfile(self.test_layout_path(), paths.hardware_layout)
        # Use an instrumented proactor
        sub_types = app_type.make_subtypes()
        sub_types.proactor_type = InstrumentedProactor
        app_settings = app_type.update_settings_from_command_line(
            app_type.get_settings(settings=app_settings, paths=paths),
            verbose=self.verbose or app_verbose,
            message_summary=self.message_summary or app_message_summary,
        )
        if not self.lifecycle_logging and not self.verbose and not app_verbose:
            app_settings.logging.levels.lifecycle = logging.WARNING
        if app_settings.logging.base_log_name == str(DEFAULT_LOG_BASE_NAME):
            app_settings.logging.base_log_name = f"{DEFAULT_LOG_BASE_NAME}-{paths.name}"

        # Create the app
        app = app_type(
            paths=paths,
            app_settings=app_settings
            if app_settings is None
            else app_settings.with_paths(paths=paths),
            sub_types=sub_types,
            layout=layout,
        )
        # Copy keys.
        if uses_tls(app.config.settings):
            copy_keys(
                str(app.config.settings.paths.name),
                app.config.settings,
            )
        return app

    @property
    def parent(self) -> InstrumentedProactor:
        if self.parent_app.proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.parent accessed before creating parent."
                "pass add_parent=True to CommTestHelper constructor or call "
                "CommTestHelper.add_parent()"
            )
        return typing.cast(InstrumentedProactor, self.parent_app.proactor)

    @property
    def child(self) -> InstrumentedProactor:
        if self.child_app.raw_proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.child accessed before creating child."
                "pass add_child=True to CommTestHelper constructor or call "
                "CommTestHelper.add_child()"
            )
        return typing.cast(InstrumentedProactor, self.child_app.proactor)

    def start_child(
        self,
    ) -> Self:
        if self.child_app.raw_proactor is None:
            self.add_child()
        return self.start_proactor(self.child)

    def start_parent(
        self,
    ) -> Self:
        if self.parent_app.raw_proactor is None:
            self.add_parent()
        return self.start_proactor(self.parent)

    def start_proactor(self, proactor: Proactor) -> Self:
        asyncio.create_task(proactor.run_forever(), name=f"{proactor.name}_run_forever")  # noqa: RUF006
        return self

    def start(
        self,
    ) -> Self:
        return self

    def add_child(
        self,
    ) -> Self:
        self.child_app.instantiate()
        return self

    def add_parent(
        self,
    ) -> Self:
        self.parent_app.instantiate()
        return self

    def remove_child(
        self,
    ) -> Self:
        self.child_app.raw_proactor = None
        return self

    def remove_parent(
        self,
    ) -> Self:
        self.parent_app.raw_proactor = None
        return self

    @classmethod
    def _get_clients_supporting_tls(cls, settings: BaseSettings) -> list[MQTTClient]:
        clients = []
        for field_name in settings.__pydantic_fields__:
            v = getattr(settings, field_name)
            if isinstance(v, MQTTClient):
                clients.append(v)
        return clients

    def _get_child_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ChildSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.child_app.config.settings)

    def _get_parent_clients_supporting_tls(self) -> list[MQTTClient]:
        """Overide to filter which MQTT clients of ParentSettingsT are treated as supporting TLS"""
        return self._get_clients_supporting_tls(self.parent_app.config.settings)

    @classmethod
    def _set_settings_use_tls(cls, use_tls: bool, clients: list[MQTTClient]) -> None:
        for client in clients:
            client.tls.use_tls = use_tls

    def set_use_tls(self, use_tls: bool) -> None:
        """Set MQTTClients which support TLS in parent and child settings to use TLS per use_tls. Clients supporting TLS
        is determined by _get_child_clients_supporting_tls() and _get_parent_clients_supporting_tls() which may be
        overriden in derived class.
        """
        self._set_settings_use_tls(use_tls, self._get_child_clients_supporting_tls())
        self._set_settings_use_tls(use_tls, self._get_parent_clients_supporting_tls())

    def setup_logging(self) -> None:
        child_settings = self.child_app.config.settings
        parent_settings = self.parent_app.config.settings
        child_settings.paths.mkdirs(parents=True)
        parent_settings.paths.mkdirs(parents=True)
        errors: list[Exception] = []
        if not self.lifecycle_logging and not self.verbose:
            if not self.child_verbose:
                child_settings.logging.levels.lifecycle = logging.WARNING
            if not self.parent_verbose:
                parent_settings.logging.levels.lifecycle = logging.WARNING
        self.logger_guards = LoggerGuards(
            list(child_settings.logging.qualified_logger_names().values())
            + list(parent_settings.logging.qualified_logger_names().values())
        )
        setup_logging(
            child_settings,
            errors=errors,
            add_screen_handler=True,
            root_gets_handlers=False,
        )
        assert not errors
        setup_logging(
            parent_settings,
            errors=errors,
            add_screen_handler=self.parent_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[InstrumentedProactor]:
        proactors = []
        if self.child_app.raw_proactor is not None:
            proactors.append(self.child)
        if self.parent_app.raw_proactor is not None:
            proactors.append(self.parent)
        return proactors

    async def stop_and_join(self) -> None:
        proactors = self.get_proactors()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                proactor.stop()
        for proactor in proactors:
            with contextlib.suppress(Exception):
                await proactor.join()

    async def __aenter__(
        self,
    ) -> Self:
        return self

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"\nCommTestHelper caught error:\n"
            f"\t{exc}.\n"
            f"\tTime: {datetime.datetime.now()}\n"  # noqa: DTZ005
            "Working log dirs:"
            f"\n\t{self.child_app.config.settings.paths.log_dir}"
            f"\n\t{self.parent_app.config.settings.paths.log_dir}"
        )

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> bool:  # noqa
        try:
            await self.stop_and_join()
        finally:
            if exc is not None:
                try:
                    s = self.get_log_path_str(exc)
                except Exception as e:  # noqa: BLE001
                    try:
                        s = (
                            f"Caught {type(e)} / <{e}> while logging "
                            f"{type(exc)} / <{exc}>"
                        )
                    except:  # noqa: E722
                        s = "ERRORs upon errors in CommTestHelper cleanup"
                with contextlib.suppress(Exception):
                    logging.getLogger("gridworks").error(s)
            with contextlib.suppress(Exception):
                self.logger_guards.restore()  # noqa
        return False

    def summary_str(self) -> str:
        s = ""
        if self.child_app.raw_proactor is not None:
            s += "CHILD:\n" f"{self.child.summary_str()}\n"
        else:
            s += "CHILD: None\n"
        if self.parent_app.raw_proactor is not None:
            s += "PARENT:\n" f"{self.parent.summary_str()}"
        else:
            s += "PARENT: None\n"
        return s

    async def await_for(
        self,
        f: Predicate | AwaitablePredicate,
        tag: str = "",
        *,
        timeout: float = 10.0,  # noqa: ASYNC109
        raise_timeout: bool = True,
        retry_duration: float = 0.01,
        err_str_f: Optional[ErrorStringFunction] = None,
        logger: Optional[logging.Logger | logging.LoggerAdapter[logging.Logger]] = None,
        error_dict: Optional[dict[str, typing.Any]] = None,
        caller_depth: int = 2,
    ) -> bool:
        if not isinstance(tag, str):
            raise TypeError(
                "ERROR. LiveTest.await_for() received a non-string tag "
                f"(type: {type(tag)}).\n"
                "  Did you pass the timeout as the second parameter?\n\n"
                "  The signature of LiveTest differs from wait.await_for(), "
                "which has timeout as the second, not third parameter."
            )
        if err_str_f is None:
            err_str_f = self.summary_str
        return await await_for(
            f=f,
            timeout=timeout,
            tag=tag,
            raise_timeout=raise_timeout,
            retry_duration=retry_duration,
            err_str_f=err_str_f,
            logger=logger,
            error_dict=error_dict,
            caller_depth=caller_depth,
        )

    async def child_to_parent_active(self) -> bool:
        return await self.await_for(
            self.child.links.link(self.child.upstream_client).active,
            "ERROR waiting child to parent link to be active",
            caller_depth=3,
        )

    async def parent_to_child_active(self) -> bool:
        return await self.await_for(
            self.parent.links.link(self.parent.downstream_client).active,
            "ERROR waiting child to parent link to be active",
            caller_depth=3,
        )

    def assert_child_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_child1_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_parent_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_acks_consistent(
        self,
        *,
        print_summary: bool = False,
        verbose: bool = False,
        log_level: int = logging.ERROR,
        raise_errors: bool = True,
    ) -> None:
        called_from_str = (
            f"\nassert_acks_consistent() called from {caller_str(depth=2)}"
        )
        counts = EventAckCounts(parent=self.parent, child=self.child, verbose=verbose)
        if not counts.ok() and raise_errors:
            raise AssertionError(
                f"ERROR {called_from_str}\n{counts.report}\n{self.summary_str()}"
            )
        if verbose or not counts.ok():
            self.child.logger.log(log_level, f"{called_from_str}\n{counts.report}")
        elif print_summary:
            self.child.logger.log(log_level, f"{called_from_str}\n{counts.summary}")

    async def await_child_at_rest(
        self,
        *,
        exp_child_persists: Optional[int | RangeTuple] = None,  # noqa: ARG002
        exact: bool = False,  # noqa: ARG002
        caller_depth: int = 4,
    ) -> None:
        child = self.child
        child_link = child.links.link(child.upstream_client)
        # Multiple waits for clarity when something goes wrong, rather than
        # one long wait with many possible failures.
        await self.await_for(
            lambda: child_link.active(),
            "ERROR in await_quiescent_connections: waiting for child link to be active",
            caller_depth=3,
        )
        await self.await_for(
            lambda: child.events_at_rest(),
            "ERROR in await_quiescent_connections: waiting for child events to upload",
            caller_depth=caller_depth,
        )

    async def await_parent_at_rest(
        self,
        *,
        exp_parent_pending: int | MinRangeTuple,
        exp_parent_persists: Optional[int | MinRangeTuple] = None,
        exact: bool = False,
        caller_depth: int = 4,
    ) -> None:
        parent = self.parent
        parent_link = parent.links.link(parent.downstream_client)
        await self.await_for(
            lambda: parent_link.active(),
            "ERROR in await_quiescent_connections: waiting for parent link to be active",
            caller_depth=3,
        )
        num_pending = range_min(exp_parent_pending)
        await self.await_for(
            lambda: parent.events_at_rest(
                num_pending=num_pending,
                exact_pending=exact,
                num_persists=None
                if exp_parent_persists is None
                else range_min(exp_parent_persists),
                exact_persists=exact,
            ),
            f"ERROR in await_quiescent_connections: waiting for parent to persist {num_pending} events",
            caller_depth=caller_depth,
        )

    # noinspection PyMethodMayBeStatic
    def default_quiesecent_child_persists(self) -> int:
        return sum(
            [
                1,  # child startup
                2,  # child connect, subscribe
            ]
        )

    def default_quiesecent_parent_pending(
        self,
        exp_child_persists: Optional[int | MinRangeTuple] = None,
    ) -> int:
        return sum(
            [
                range_min(
                    exp_child_persists
                    if exp_child_persists is not None
                    else self.default_quiesecent_child_persists()
                ),
                1,  # child peer active
                1,  # parent startup
                3,  # parent connect, subscribe, peer active
            ]
        )

    def default_quiesecent_parent_persists(
        self,
        exp_parent_pending: Optional[int | MinRangeTuple] = None,
        exp_child_persists: Optional[int | MinRangeTuple] = None,
    ) -> int:
        return range_min(
            exp_parent_pending
            if exp_parent_pending is not None
            else self.default_quiesecent_parent_pending(
                exp_child_persists=exp_child_persists
            )
        )

    def assert_quiescent_event_counts(
        self,
        *,
        exp_child_persists: int | tuple[int | None, int | None],
        exp_parent_pending: int | tuple[int | None, int | None],
        exp_parent_persists: int | tuple[int | None, int | None],
        exact: bool = False,
    ) -> None:
        exp_child_persists = as_range_tuple(exp_child_persists, exact)
        exp_parent_pending = as_range_tuple(exp_parent_pending, exact)
        exp_parent_persists = as_range_tuple(exp_parent_persists, exact)
        self.parent.assert_event_counts(
            num_pending=exp_parent_pending,
            num_persists=exp_parent_persists,
            num_clears=0,
            num_retrieves=0,
            tag="parent",
            err_str=self.summary_str(),
        )
        self.child.assert_event_counts(
            num_persists=exp_child_persists,
            all_clear=True,
            tag="child",
            err_str=self.summary_str(),
        )

    @classmethod
    def _as_tuple(
        cls,
        val: Optional[int | RangeTuple],
        default_func: typing.Callable[[], int],
        exact: bool,
    ) -> MinRangeTuple:
        return (
            as_min_range_tuple(default_func(), exact=exact)
            if val is None
            else as_min_range_tuple(val, exact=exact)
        )

    async def await_quiescent_connections(
        self,
        *,
        exp_child_persists: Optional[int | RangeTuple] = None,
        exp_parent_pending: Optional[int | RangeTuple] = None,
        exp_parent_persists: Optional[int | RangeTuple] = None,
        exact: bool = False,
        assert_event_counts: bool = True,
    ) -> None:
        exp_child_persists = self._as_tuple(
            exp_child_persists, self.default_quiesecent_child_persists, exact
        )
        exp_parent_pending = self._as_tuple(
            exp_parent_pending,
            functools.partial(
                self.default_quiesecent_parent_pending, exp_child_persists[0]
            ),
            exact,
        )
        exp_parent_persists = self._as_tuple(
            exp_parent_persists,
            functools.partial(
                self.default_quiesecent_parent_persists,
                exp_parent_pending=exp_child_persists,
                exp_child_persists=exp_child_persists,
            ),
            exact,
        )
        # Multiple waits for clarity when something goes wrong, rather than
        # one long wait with many possible failures.
        await self.await_child_at_rest(
            exp_child_persists=exp_child_persists,
            exact=exact,
        )
        await self.await_parent_at_rest(
            exp_parent_pending=exp_parent_pending,
            exact=exact,
            exp_parent_persists=exp_parent_persists,
        )
        if assert_event_counts:
            self.assert_quiescent_event_counts(
                exp_child_persists=exp_child_persists,
                exp_parent_pending=exp_parent_pending,
                exp_parent_persists=exp_parent_persists,
                exact=exact,
            )

    def pings_from_parent_topic(self) -> str:
        return MQTTTopic.encode(
            envelope_type="gw",
            src=self.parent.publication_name,
            dst=self.parent.links.topic_dst(self.parent.downstream_client),
            message_type="gridworks-ping",
        )

    def pings_from_child_topic(self) -> str:
        return MQTTTopic.encode(
            envelope_type="gw",
            src=self.child.publication_name,
            dst=self.child.links.topic_dst(self.child.downstream_client),
            message_type="gridworks-ping",
        )

    def pings_from_parent(self) -> int:
        return self.child.stats.link(self.child.upstream_client).num_received_by_topic[
            self.pings_from_parent_topic()
        ]

    def pings_from_child(self) -> int:
        return self.parent.stats.link(
            self.parent.downstream_client
        ).num_received_by_topic[self.pings_from_child_topic()]

    def child_comm_events(self) -> list[CommEvent]:
        return self.child.stats.link(self.child.upstream_client).comm_events

    def parent_comm_events(self) -> list[CommEvent]:
        return self.parent.stats.link(self.parent.downstream_client).comm_events



import datetime
import textwrap

from gwproto.messages import Ack, AnyEvent, EventBase
from result import Err, Ok

from gwproactor import Problems
from gwproactor.persister import ByteDecodingError, UIDMissingWarning
from gwproactor_test.instrumented_proactor import InstrumentedProactor


class EventAckConsistencyError(Exception): ...


class UnexpectedMessage(EventAckConsistencyError): ...


class CountInconsistency(EventAckConsistencyError): ...


class _EventAckCountsIntermediate:
    problems: Problems

    paused_ack_list: list[Ack]

    in_flight_set: set[str]
    pending_set: set[str]
    paused_ack_set: set[str]
    in_flight_not_paused_set: set[str]
    pending_not_paused_set: set[str]
    paused_not_events_set: set[str]

    pending_events: dict[str, EventBase]
    in_flight_not_paused_events: dict[str, EventBase]
    pending_not_paused_events: dict[str, EventBase]
    paused_not_events_list: list[str]

    def __init__(
        self,
        *,
        parent: InstrumentedProactor,
        child: InstrumentedProactor,
        verbose: bool = False,
    ) -> None:
        self._find_errors(parent=parent, child=child)
        if self.ok() and not verbose:
            self.pending_events = {}
            self.in_flight_not_paused_events = {}
            self.pending_not_paused_events = {}
            self.paused_not_events_list = []
        else:
            self._sort_events(child=child)

    def _find_errors(
        self, *, parent: InstrumentedProactor, child: InstrumentedProactor
    ) -> None:
        in_flight_set = set(child.links.in_flight_events.keys())
        pending_set = set(child.event_persister.pending_ids())
        paused_ack_list: list[Ack] = []
        problems = Problems()
        for paused in parent.needs_ack:
            if paused.link_name == parent.downstream_client:
                match paused.message.Payload:
                    case Ack():
                        paused_ack_list.append(paused.message.Payload)
                    case _:
                        problems.add_error(
                            UnexpectedMessage(paused.message.Header.TypeName)
                        )
        paused_ack_set = {paused_ack.AckMessageID for paused_ack in paused_ack_list}
        in_flight_not_paused_set = in_flight_set - paused_ack_set
        pending_not_paused_set = pending_set - paused_ack_set
        paused_not_events_set = paused_ack_set - (in_flight_set | paused_ack_set)
        self.problems = problems
        self.in_flight_set = in_flight_set
        self.pending_set = pending_set
        self.paused_ack_list = paused_ack_list
        self.paused_ack_set = paused_ack_set
        self.in_flight_not_paused_set = in_flight_not_paused_set
        self.pending_not_paused_set = pending_not_paused_set
        self.paused_not_events_set = paused_not_events_set

    def _sort_pending_events(self, *, child: InstrumentedProactor) -> None:
        # Sort pending events by time (persister does not guarantee order)
        pending_event_list: list[AnyEvent] = []
        for event_id in self.pending_set:
            match child.event_persister.retrieve(event_id):
                case Ok(content):
                    if content is not None:
                        try:
                            pending_event_list.append(
                                AnyEvent.model_validate_json(content)
                            )
                        except Exception as e:  # noqa: BLE001
                            self.problems.add_error(e).add_error(
                                ByteDecodingError("reupload_events", uid=event_id)
                            )
                    else:
                        self.problems.add_error(
                            UIDMissingWarning("Ack consistency check", uid=event_id)
                        )
                case Err(one_retrieve_problems):
                    self.problems.add_error(one_retrieve_problems)
        child.event_persister._num_retrieves -= len(pending_event_list)  # type: ignore # noqa
        pending_event_list.sort(key=lambda event_: event_.TimeCreatedMs)
        self.pending_events = {
            pending_event.MessageId: pending_event
            for pending_event in pending_event_list
        }

    def _sort_events(self, *, child: InstrumentedProactor) -> None:
        self._sort_pending_events(child=child)
        self.in_flight_not_paused_events = {
            event.MessageId: event
            for event in sorted(
                [
                    child.links.in_flight_events[x]
                    for x in self.in_flight_not_paused_set
                ],
                key=lambda x: x.TimeCreatedMs,
            )
        }
        self.pending_not_paused_events = {
            event.MessageId: event
            for event in sorted(
                [self.pending_events[x] for x in self.pending_not_paused_set],
                key=lambda x: x.TimeCreatedMs,
            )
        }
        self.paused_not_events_list = [
            ack.AckMessageID
            for ack in self.paused_ack_list
            if ack.AckMessageID in self.paused_not_events_set
        ]

    def ok(self) -> bool:
        return (
            not self.problems
            and not self.in_flight_not_paused_set
            and not self.pending_not_paused_set
            and not self.paused_not_events_set
        )

    def __bool__(self) -> bool:
        return self.ok()


class _EventAckReportGenerator:
    child: InstrumentedProactor
    c: _EventAckCountsIntermediate
    verbose: bool
    summary: str
    non_error_report: str
    error_report: str
    report: str

    def __init__(
        self,
        child: InstrumentedProactor,
        c: _EventAckCountsIntermediate,
        *,
        verbose: bool = False,
    ) -> None:
        self.child = child
        self.c = c
        self.verbose = verbose
        self._summarize()
        if not verbose and self.c.ok():
            self.summary += "Acks CONSISTENT\n"
            self.report = self.summary
        else:
            self._make_non_error_report()
            self._make_error_report()
            if self.c.ok():
                self.summary += "Acks CONSISTENT\n"
            self.report = self.non_error_report + self.error_report + self.summary

    def _summarize(self) -> None:
        self.summary = (
            f"Parent paused acks: {len(self.c.paused_ack_set)}\n"
            f"Child pending events: {len(self.c.pending_set)}\n"
            f"Child in-flight events: {len(self.c.in_flight_set)}\n"
        )

    def _make_non_error_report(self) -> None:
        report = f"Parent paused acks: {len(self.c.paused_ack_list)}\n"
        event: EventBase | str
        for i, paused_ack in enumerate(self.c.paused_ack_list):
            if paused_ack.AckMessageID in self.child.links.in_flight_events:
                event = self.child.links.in_flight_events[paused_ack.AckMessageID]
                loc = "in-flight"
            elif paused_ack.AckMessageID in self.c.pending_events:
                event = self.c.pending_events[paused_ack.AckMessageID]
                loc = "pending"
            else:
                event = paused_ack.AckMessageID
                loc = "*UKNONWN*"
            report += self._event_line(event, loc, i + 1, len(self.c.paused_ack_list))
        report += f"Child in-flight events: {self.child.links.num_in_flight}\n"
        for i, event in enumerate(self.child.links.in_flight_events.values()):
            report += self._event_line(
                event, "in-flight", i + 1, len(self.c.pending_events)
            )
        report += f"Child pending events: {len(self.c.pending_events)}\n"
        for i, event in enumerate(self.c.pending_events.values()):
            report += self._event_line(
                event, "pending", i + 1, len(self.c.pending_events)
            )
        self.non_error_report = report

    def _make_error_report(self) -> None:
        report = ""
        events = self.c.in_flight_not_paused_events.values()
        line = f"Child in-flight events not in parent paused acks: {len(events)}\n"
        report += line
        self.summary += line
        for i, event in enumerate(events):
            report += self._event_line(event, "in-flight", i + 1, len(events))

        events = self.c.pending_not_paused_events.values()
        line = f"Child pending events not in parent paused acks: {len(events)}\n"
        report += line
        self.summary += line
        for i, event in enumerate(events):
            report += self._event_line(event, "pending", i + 1, len(events))

        event_ids = self.c.paused_not_events_list
        line = f"Paused acks not in child in-flight or pending: {len(event_ids)}\n"
        report += line
        self.summary += line
        for i, event_id in enumerate(event_ids):
            report += self._event_line(event_id, "", i + 1, len(event_ids))

        line = f"Problems making ack count calculation: {len(self.c.problems)}"
        report += line
        self.summary += line
        report += textwrap.indent(str(self.c.problems), "  ")

        self.error_report = report

    @classmethod
    def _event_line(cls, event: EventBase | str, loc: str, i: int, n: int) -> str:
        if isinstance(event, EventBase):
            event_id = event.MessageId
            dt = datetime.datetime.fromtimestamp(
                event.TimeCreatedMs / 1000, tz=datetime.UTC
            )
            info_s = f"{dt}   {event.TypeName}"
        else:
            event_id = event
            info_s = ""
        return f"  {i:3d} / {n:3d}   {event_id[:8]}   {loc:10s}   {info_s}\n"

    @property
    def problems(self) -> Problems:
        return self.c.problems


class EventAckCounts:
    summary: str = ""
    report: str = ""
    problems: Problems

    def __init__(
        self,
        *,
        parent: InstrumentedProactor,
        child: InstrumentedProactor,
        verbose: bool = False,
    ) -> None:
        calc = _EventAckCountsIntermediate(parent=parent, child=child, verbose=verbose)
        reporter = _EventAckReportGenerator(child=child, c=calc, verbose=verbose)
        self.problems = calc.problems
        self.summary = reporter.summary
        self.report = reporter.report

    def ok(self) -> bool:
        return not self.problems

    def __bool__(self) -> bool:
        return self.ok()



import functools
import logging
import textwrap
import typing
from typing import Any, Optional, Self

from gwproto import HardwareLayout

from gwproactor import AppSettings, setup_logging
from gwproactor.app import App
from gwproactor.config import MQTTClient
from gwproactor_test.dummies.tree.atn import DummyAtnApp
from gwproactor_test.dummies.tree.scada1 import DummyScada1App
from gwproactor_test.dummies.tree.scada2 import DummyScada2App
from gwproactor_test.event_consistency_checks import EventAckCounts
from gwproactor_test.instrumented_proactor import (
    InstrumentedProactor,
    MinRangeTuple,
    RangeTuple,
    as_range_tuple,
    caller_str,
)
from gwproactor_test.live_test_helper import (
    LiveTest,
    get_option_value,
)
from gwproactor_test.logger_guard import LoggerGuards


class TreeLiveTest(LiveTest):
    _child2_app: App
    child2_verbose: bool = False
    child2_message_summary: bool = False
    child2_on_screen: bool = False
    child2_logger_guards: LoggerGuards

    def __init__(
        self,
        *,
        add_child1: bool = False,
        start_child1: bool = False,
        child1_verbose: Optional[bool] = None,
        child1_message_summary: Optional[bool] = None,
        child1_layout: Optional[HardwareLayout] = None,
        child2_app_settings: Optional[AppSettings] = None,
        child2_layout: Optional[HardwareLayout] = None,
        child2_verbose: Optional[bool] = None,
        child2_message_summary: Optional[bool] = None,
        add_child2: bool = False,
        start_child2: bool = False,
        child2_on_screen: Optional[bool] = None,
        **kwargs: Any,
    ) -> None:
        kwargs["add_child"] = add_child1 or kwargs.get("add_child", False)
        kwargs["start_child"] = start_child1 or kwargs.get("start_child", False)
        if child1_verbose is None:
            kwargs["child_verbose"] = get_option_value(
                parameter_value=child1_verbose,
                option_name="--child1-verbose",
                request=kwargs.get("request"),
            )
        if child1_message_summary is None:
            kwargs["child_message_summary"] = get_option_value(
                parameter_value=child1_verbose,
                option_name="--child1-message-summary",
                request=kwargs.get("request"),
            )
        kwargs["child_layout"] = child1_layout or kwargs.get("child_layout")
        kwargs["request"] = kwargs.get("request")
        super().__init__(**kwargs)
        self.child2_verbose = get_option_value(
            parameter_value=child2_verbose,
            option_name="--child2-verbose",
            request=kwargs.get("request"),
        )
        self.child2_message_summary = get_option_value(
            parameter_value=child2_message_summary,
            option_name="--child2-message-summary",
            request=kwargs.get("request"),
        )
        self.child2_on_screen = get_option_value(
            parameter_value=child2_on_screen,
            option_name="--child2-on-screen",
            request=kwargs.get("request"),
        )
        self._child2_app = self._make_app(
            self.child2_app_type(),
            child2_app_settings,
            app_verbose=self.child2_verbose,
            app_message_summary=self.child2_message_summary,
            layout=child2_layout if child2_layout is not None else kwargs.get("layout"),
        )
        self.setup_child2_logging()
        add_child2 = (
            add_child2
            or start_child2
            or kwargs.get("add_all", False)
            or kwargs.get("start_all", False)
        )
        start_child2 = start_child2 or kwargs.get("start_all", False)
        if add_child2 or start_child2:
            self.add_child2()
            if start_child2:
                self.start_child2()

    @classmethod
    def child_app_type(cls) -> type[App]:
        return DummyScada1App

    @property
    def child_app(self) -> DummyScada1App:
        return typing.cast(DummyScada1App, self._child_app)

    @property
    def child1_app(self) -> DummyScada1App:
        return self.child_app

    @classmethod
    def child2_app_type(cls) -> type[App]:
        return DummyScada2App

    @property
    def child2_app(self) -> DummyScada2App:
        return typing.cast(DummyScada2App, self._child2_app)

    @classmethod
    def parent_app_type(cls) -> type[App]:
        return DummyAtnApp

    @property
    def parent_app(self) -> DummyAtnApp:
        return typing.cast(DummyAtnApp, self._parent_app)

    @property
    def child1(self) -> InstrumentedProactor:
        return self.child

    def add_child1(self) -> Self:
        return self.add_child()

    def start_child1(
        self,
    ) -> Self:
        return self.start_child()

    def remove_child1(
        self,
    ) -> Self:
        return self.remove_child()

    @property
    def child2(self) -> InstrumentedProactor:
        if self.child2_app.proactor is None:
            raise RuntimeError(
                "ERROR. CommTestHelper.child accessed before creating child."
                "pass add_child=True to CommTestHelper constructor or call "
                "CommTestHelper.add_child()"
            )
        return typing.cast(InstrumentedProactor, self.child2_app.proactor)

    def add_child2(
        self,
    ) -> Self:
        self.child2_app.instantiate()
        return self

    def start_child2(
        self,
    ) -> Self:
        if self.child2_app.raw_proactor is None:
            self.add_child2()
        return self.start_proactor(self.child2)

    def remove_child2(
        self,
    ) -> Self:
        self.child2_app.raw_proactor = None
        return self

    def _get_child2_clients_supporting_tls(self) -> list[MQTTClient]:
        return self._get_clients_supporting_tls(self.child2_app.config.settings)

    def set_use_tls(self, use_tls: bool) -> None:
        super().set_use_tls(use_tls)
        self._set_settings_use_tls(use_tls, self._get_child2_clients_supporting_tls())

    def setup_child2_logging(self) -> None:
        self.child2_app.config.settings.paths.mkdirs(parents=True)
        errors: list[Exception] = []
        self.logger_guards.add_loggers(
            list(
                self.child2_app.config.settings.logging.qualified_logger_names().values()
            )
        )
        setup_logging(
            self.child2_app.config.settings,
            errors=errors,
            add_screen_handler=self.child2_on_screen,
            root_gets_handlers=False,
        )
        assert not errors

    def get_proactors(self) -> list[InstrumentedProactor]:
        proactors = super().get_proactors()
        if self.child2_app.raw_proactor is not None:
            proactors.append(self.child2)
        return proactors

    def get_log_path_str(self, exc: BaseException) -> str:
        return (
            f"CommTestHelper caught error {exc}.\n"
            "Working log dirs:"
            f"\n\t[{self.child_app.config.settings.paths.log_dir}]"
            f"\n\t[{self.parent_app.config.settings.paths.log_dir}]"
        )

    def summary_str(self) -> str:
        s = ""
        if self.child_app.raw_proactor is None:
            s += "SCADA1: None\n"
        else:
            s += "SCADA1:\n"
            s += textwrap.indent(self.child1.summary_str(), "    ") + "\n"
        if self.child2_app.raw_proactor is None:
            s += "SCADA2: None\n"
        else:
            s += "SCADA2:\n"
            s += textwrap.indent(self.child2.summary_str(), "    ") + "\n"
        if self.parent_app.raw_proactor is None:
            s += "ATN: None\n"
        else:
            s += "ATN:\n"
            s += textwrap.indent(self.parent.summary_str(), "    ") + "\n"
        return s

    def assert_child1_events_at_rest(
        self, *args: typing.Any, **kwargs: typing.Any
    ) -> None:
        self.child.assert_event_counts(*args, **kwargs)

    def assert_acks_consistent(
        self,
        *,
        print_summary: bool = False,
        verbose: bool = False,
        log_level: int = logging.ERROR,
        raise_errors: bool = True,
        child1_to_parent: bool = True,
        child2_to_child1: bool = True,
    ) -> None:
        called_from_str = (
            f"\nassert_acks_consistent() called from {caller_str(depth=2)}"
        )
        counts_ok = True
        summary_str = ""
        report_str = ""
        for check_pair, parent, child, prefix in [
            (child1_to_parent, self.parent, self.child, "Child1 / Parent"),
            (child2_to_child1, self.child, self.child2, "Child2 / Child1"),
        ]:
            if check_pair:
                counts = EventAckCounts(parent=parent, child=child, verbose=verbose)
                counts_ok = counts_ok and counts.ok()
                summary_str += f"{prefix} summary\n{counts.summary}\n"
                report_str += f"{prefix} report\n{counts.report}\n"
        if not counts_ok and raise_errors:
            raise AssertionError(f"ERROR {called_from_str}\n{report_str}")
        if verbose or not counts_ok:
            self.child.logger.log(log_level, f"{called_from_str}\n{report_str}")
        elif print_summary:
            self.child.logger.log(log_level, f"{called_from_str}\n{summary_str}")

    async def await_child2_at_rest(
        self,
        *,
        exp_child_persists: Optional[int] = None,  # noqa: ARG002
        exact: bool = False,  # noqa: ARG002
        caller_depth: int = 4,
    ) -> None:
        # Multiple waits for clarity when something goes wrong, rather than
        # one long wait with many possible failures.
        await self.await_for(
            lambda: self.child2.links.link(self.child2.upstream_client).active(),
            (
                "ERROR in await_quiescent_connections: waiting for child2 to "
                "child1 link to be active"
            ),
            caller_depth=caller_depth,
        )
        await self.await_for(
            lambda: self.child2.events_at_rest(),
            (
                "ERROR in await_quiescent_connections: waiting for child2 "
                "events to upload to child1"
            ),
            caller_depth=caller_depth,
        )

    async def await_child_at_rest(
        self,
        *,
        exp_child_persists: Optional[int | RangeTuple] = None,
        exact: bool = False,
        caller_depth: int = 5,
    ) -> None:
        await self.await_child1_at_rest(
            exp_child_persists=exp_child_persists,
            exact=exact,
            caller_depth=caller_depth,
        )

    async def await_child1_at_rest(
        self,
        *,
        exp_child_persists: Optional[int | RangeTuple] = None,  # noqa: ARG002
        exact: bool = False,  # noqa: ARG002
        caller_depth: int = 4,
    ) -> None:
        await self.await_for(
            lambda: self.child1.links.link(self.child1.downstream_client).active(),
            "ERROR in await_quiescent_connections: waiting for child1 to child2 link to be active",
            caller_depth=caller_depth,
        )
        await self.await_for(
            lambda: self.child1.links.link(self.child1.upstream_client).active(),
            "ERROR in await_quiescent_connections: waiting for child1 to parent link to be active",
            caller_depth=caller_depth,
        )
        await self.await_for(
            lambda: self.child1.events_at_rest(),
            "ERROR in await_quiescent_connections: waiting for child1 events to upload to parent",
            caller_depth=caller_depth,
        )

    async def await_parent_at_rest(
        self,
        *,
        exp_parent_pending: int | MinRangeTuple,
        exp_parent_persists: Optional[int | MinRangeTuple] = None,
        exp_total_children_events: Optional[int] = None,  # noqa: ARG002
        exact: bool = False,
        caller_depth: int = 4,
    ) -> None:
        await super().await_parent_at_rest(
            exp_parent_pending=exp_parent_pending,
            exp_parent_persists=exp_parent_persists,
            exact=exact,
            caller_depth=caller_depth + 1,
        )

    # noinspection PyMethodMayBeStatic
    def default_quiesecent_child2_persists(self) -> int:
        # child2 will persist at least 3 events (startup, connect, subscribe),
        # but it could persist more depending on when parent link went active
        # w.r.t the admin and child1. This is not, however, predictable, since,
        # especially in test runs with many tests, mqtt thrash is common,
        # resulting in more events persisted.
        return 3

    # noinspection PyMethodMayBeStatic
    def default_quiescent_total_children_events(self) -> int:
        return sum(
            [
                1,  # child2 startup
                4,  # child2 (child1, admin) x (connect, substribe)
                1,  # child2 peer active
                1,  # child1 startup
                6,  # child1 (parent, child2, admin) x (connect, subscribe)
                2,  # child1 (parent, child2) x peer active
            ]
        )

    def default_quiesecent_child_persists(self) -> int:
        return self.default_quiesecent_child1_persists()

    # noinspection PyMethodMayBeStatic
    def default_quiesecent_child1_persists(self) -> int:
        # child1 will persist at least 3 events (startup, connect, subscribe),
        # but it could persist more depending on when parent link went active
        # w.r.t the admin and child2. This is not, however, predictable, since,
        # especially in test runs with many tests, mqtt thrash is common,
        # resulting in more events persisted.
        return 3

    def default_quiesecent_parent_pending(
        self,
        exp_child_persists: Optional[int | MinRangeTuple] = None,
        exp_total_children_events: Optional[int] = None,
    ) -> int:
        if exp_child_persists is not None:
            raise ValueError(
                "ERROR. TreeLiveTestHelper.default_quiesecent_child_persists() "
                "does not support exp_child_persists because it needs a count "
                "from both childre. Use exp_total_children_events instead."
            )
        return sum(
            [
                exp_total_children_events
                if exp_total_children_events is not None
                else self.default_quiescent_total_children_events(),
                1,  # parent startup
                2,  # parent connect, subscribe
                1,  # child1 peer active
            ]
        )

    def assert_quiescent_tree_event_counts(
        self,
        *,
        exp_child1_persists: int | RangeTuple,
        exp_child2_persists: int | RangeTuple,
        exp_parent_pending: int | RangeTuple,
        exp_parent_persists: int | RangeTuple,
        exact: bool = False,
    ) -> None:
        exp_child1_persists = as_range_tuple(exp_child1_persists, exact)
        exp_child2_persists = as_range_tuple(exp_child2_persists, exact)
        exp_parent_pending = as_range_tuple(exp_parent_pending, exact)
        exp_parent_persists = as_range_tuple(exp_parent_persists, exact)
        summary_str = self.summary_str()
        self.child2.assert_event_counts(
            num_persists=exp_child2_persists,
            all_clear=True,
            tag="child2",
            err_str=summary_str,
        )
        self.child1.assert_event_counts(
            num_persists=exp_child1_persists,
            all_clear=True,
            tag="child1",
            err_str=summary_str,
        )
        self.parent.assert_event_counts(
            num_pending=exp_parent_pending,
            num_persists=exp_parent_persists,
            num_clears=0,
            num_retrieves=0,
            tag="parent",
            err_str=summary_str,
        )

    async def await_quiescent_connections(
        self,
        *,
        exp_child_persists: Optional[int | RangeTuple] = None,
        exp_child1_persists: Optional[int | RangeTuple] = None,
        exp_child2_persists: Optional[int | RangeTuple] = None,
        exp_total_children_events: Optional[int] = None,
        exp_parent_pending: Optional[int | RangeTuple] = None,
        exp_parent_persists: Optional[int | RangeTuple] = None,
        exact: bool = False,
        assert_event_counts: bool = True,
    ) -> None:
        if exp_child_persists is not None and exp_child1_persists is not None:
            raise RuntimeError(
                "Specify 0 or 1 of (exp_child_persists, exp_child1_persists)"
            )
        exp_child1_persists = self._as_tuple(
            exp_child1_persists
            if exp_child1_persists is not None
            else exp_child_persists,
            self.default_quiesecent_child1_persists,
            exact,
        )
        exp_child2_persists = self._as_tuple(
            exp_child2_persists,
            self.default_quiesecent_child2_persists,
            exact,
        )
        if exp_total_children_events is None:
            exp_total_children_events = self.default_quiescent_total_children_events()
        exp_parent_pending = self._as_tuple(
            exp_parent_pending,
            functools.partial(
                self.default_quiesecent_parent_pending,
                exp_total_children_events=exp_total_children_events
                if exp_total_children_events is not None
                else self.default_quiescent_total_children_events(),
            ),
            exact,
        )
        exp_parent_persists = self._as_tuple(
            exp_parent_persists,
            functools.partial(
                self.default_quiesecent_parent_persists,
                exp_parent_pending=exp_parent_pending,
            ),
            exact,
        )

        # Multiple waits for clarity when something goes wrong, rather than
        # one long wait with many possible failures.
        await self.await_child2_at_rest()
        await self.await_child1_at_rest()
        await self.await_parent_at_rest(
            exp_parent_pending=exp_parent_pending,
            exp_parent_persists=exp_parent_persists,
            exp_total_children_events=exp_total_children_events,
            exact=exact,
        )
        if assert_event_counts:
            self.assert_quiescent_tree_event_counts(
                exp_child1_persists=exp_child1_persists,
                exp_child2_persists=exp_child2_persists,
                exp_parent_pending=exp_parent_pending,
                exp_parent_persists=exp_parent_persists,
                exact=exact,
            )
